from abc import ABC, abstractmethod
from typing import Any, Tuple


class MigrationEngine(ABC):
    @abstractmethod
    def init_remote(self):
        ...

    @abstractmethod
    def remote_exists(self) -> bool:
        ...

    @abstractmethod
    def update_remote(self, state: str):
        ...

    @abstractmethod
    def get_remote(self) -> str:
        ...

    @abstractmethod
    def prepare_args(self) -> Tuple[Any]:
        ...



import os.path

from mako.template import Template

template_dir = os.path.dirname(__file__)
base_template = Template(filename=f"{template_dir}/base.py.mako")
spark_template = Template(filename=f"{template_dir}/spark.py.mako")
elasticsearch_template = Template(filename=f"{template_dir}/elasticsearch.py.mako")
sqlalchemy_template = Template(filename=f"{template_dir}/sqlalchemy.py.mako")



import datetime
import glob
import os
import runpy
from typing import Union

from mako.template import Template

from many.templates import base_template
from many.utils import create_hash


class Revisions:
    def __init__(
        self,
        script_location: str,
        file_template: str = "%(year)d_%(month).2d_%(day).2d_%(hour).2d%(minute).2d-%(rev)s_%(slug)s",
        truncate_slug_length: int = 40,
        template: Template = base_template,
    ):
        self.script_location = script_location
        self.file_template = file_template
        self.truncate_slug_length = truncate_slug_length
        self.template = template

    @property
    def revisions(self):
        revisions = []
        for file_path in glob.glob(self.script_location.rstrip("/") + "/*.py"):
            module = runpy.run_path(file_path)
            revisions.append(module)
        return revisions

    @property
    def up_by_current(self):
        return {k["down_version"]: k for k in self.revisions}

    @property
    def down_by_current(self):
        return {k["version"]: k for k in self.revisions}

    def has_upgrade(self, current_state: Union[str, int]):
        return current_state in self.up_by_current.keys()

    def get_upgrade(self, current_state: Union[str, int]):
        template = self.up_by_current[current_state]
        return template["up"], template["version"]

    def has_downgrade(self, current_state: Union[str, int]):
        return current_state in self.down_by_current.keys()

    def get_downgrade(self, current_state: Union[str, int]):
        template = self.down_by_current[current_state]
        return template["down"], template["down_version"]

    def get_head(self):
        for template in self.revisions[::-1]:
            version = template["version"]
            if not self.has_upgrade(version):
                return version

    def create_revision(self, m: str):
        latest = self.get_head()
        version = create_hash(m)

        content = self.template.render(version=version, down_version=latest)

        now = datetime.datetime.now()
        if not os.path.exists(self.script_location):
            os.mkdir(self.script_location)

        file_name = self.file_template % {
            "slug": m.lower().replace(" ", "_")[: self.truncate_slug_length],
            "year": now.year,
            "month": now.month,
            "day": now.day,
            "hour": now.hour,
            "minute": now.minute,
            "rev": version,
        }
        file_name = f"{self.script_location}/{file_name}.py"
        with open(file_name, "w") as file:
            file.write(content)
        print(
            f"Created new migration template with version {version} and at path"
            f" {file_name}"
        )



import os
from configparser import ConfigParser
from typing import Annotated

import typer
from typer import Option

from many.engine import MigrationEngine
from many.migrate import Migrator
from many.revise import Revisions
from many.templates import Template, base_template


def _init_revision_app(revisions: Revisions):
    app = typer.Typer()

    @app.command(help="Create new revision")
    def create(m: Annotated[str, Option("-m")]):
        revisions.create_revision(m=m)

    return app


def _init_migration_app(engine: MigrationEngine, revisions: Revisions):
    app = typer.Typer()

    migrator = Migrator(engine=engine, revisions=revisions)

    @app.command(help="Command to initialize the migrations")
    def init():
        migrator.init()

    @app.command(help="Command to upgrade the data store")
    def up(level: str = "head"):
        migrator.up(level=level)

    @app.command(help="Command to downgrade the data store")
    def down(level: str = 1):
        migrator.down(level=level)

    return app


def init_app(
    migration_engine: MigrationEngine,
    template: Template = base_template,
    config_file: str = None,
):
    conf = ConfigParser()
    conf.read(config_file or os.path.dirname(__file__) + "/default-config.ini")

    app = typer.Typer()
    revisions = Revisions(
        script_location=conf["revisions"]["script_location"],
        file_template=conf["revisions"]["file_template"],
        truncate_slug_length=int(conf["revisions"]["truncate_slug_length"]),
        template=template,
    )
    revision_app = _init_revision_app(revisions=revisions)
    migrator_app = _init_migration_app(engine=migration_engine, revisions=revisions)

    app.add_typer(revision_app, name="revision")
    app.add_typer(migrator_app, name="migrate")

    return app



import math
from typing import Union

from many.engine import MigrationEngine
from many.revise import Revisions


class Migrator:
    def __init__(self, engine: MigrationEngine, revisions: Revisions):
        self.engine = engine
        self.revisions = revisions

    def init(self):
        if self.engine.remote_exists():
            print("Remote state already exists.")
            return
        self.engine.init_remote()
        print("Remote state initialized")

    def one_up_from(self, state: Union[str, int]):
        if not self.revisions.has_upgrade(state):
            print("Migrations complete. No upgrades to run.")
            return
        method, target_state = self.revisions.get_upgrade(state)

        # Run upgrade with arguments
        method(*self.engine.prepare_args())

        # Update remote state
        self.engine.update_remote(target_state)
        print(f"Migrated successfully (up) from {state} to {target_state}.")
        return target_state

    def one_down_from(self, state: Union[str, int]):
        if not self.revisions.has_downgrade(state):
            print("Migrations complete. No downgrades to run.")
            return

        method, target_state = self.revisions.get_downgrade(state)

        # Run downgrade with arguments
        method(*self.engine.prepare_args())

        # Update remote state
        self.engine.update_remote(target_state)
        print(f"Migrated successfully (down) from {state} to {target_state}.")
        return target_state

    def get_current_state(self):
        if not self.engine.remote_exists():
            raise ValueError("Remote state is not initialized.")
        return self.engine.get_remote()

    def up(self, level: str = "head"):
        if isinstance(level, str) and level.lower() != "head" and not level.isdigit():
            raise ValueError("level must be either 'head', or an integer")
        elif not level.isdigit():
            level = math.inf
        else:
            level = int(level)

        current_state = self.get_current_state()
        i = 0
        while i < level:
            new_state = self.one_up_from(current_state)
            if new_state is None:
                break
            current_state = new_state
            i += 1

    def down(self, level: Union[int, str] = "base"):
        if isinstance(level, str) and level.lower() != "base" and not level.isdigit():
            raise ValueError("level must be either 'base', or an integer")
        elif not level.isdigit():
            level = math.inf
        else:
            level = int(level)

        current_state = self.get_current_state()
        i = 0
        while i < level:
            new_state = self.one_down_from(current_state)
            if new_state is None:
                break
            current_state = new_state
            i += 1
